=begin
Convenient logging + user information system, based on the ideas in Log4R. Some
data needs to be saved for developers or sys-admins, some data needs to be
conveyed to the user. This is closely related, since fatal stuff needs to go to
the user as well as the sys-admin. The main viewpoint here is from the user.

Whenever a GUI is used, windows might popup, but on console (tty, stdout,
stderr), plain text should be dumped and strings read so the user can intervene.

Note that p, puts and $DEBUG provide a good basis for developer information,
while software is in beta. It gets clunky when apps grow, though.

Due to the basic nature of reporting, methods here are put into the Kernel(!)
so there is no textual overhead in the code. That may be awkward for the GUI
stuff, but I explicitely want to be GUI agnostic.

Author: Kero van Gelder
Copyright: Kero van Gelder, the GNU LGPL applies
Original: derived from similar Java code, summer 1998 I think.
Since: 2002 Nov 23
Version: 2002 Nov 24
=end

module Kernel

  # Basic logger (which implements puts())

  @@Logger = $stderr

  def Kernel::set_logger(new_logger)
    @@Logger = new_logger
  end

  class SomeApp
    def status(string)
      @@Logger.puts "2 USER "+string
    end
    def dialog(string, action_thread=nil)
      @@Logger.puts string
      # if remotely possible, provide an option to call the action_thread#raise
    end
  end


  # User Informer (implementing status() and dialog())

  @@Informer = SomeApp.new()

  def Kernel::set_informer(new_informer)
    @@Informer = new_informer
  end

  # Log4r taught me this:
  #
  # Debug  close (almost paranoia) monitoring   no popup, can go to list
  # Info   regular monitoring                   to a list, might popup
  # Warn   slightly out of the ordinary         list, popup
  # Error  a problem, but can "continue"        list, popup
  # Fatal  major failure, terminates            popup (if still possible)

  # Now user dialogs are just a bit more than just logging. Popups usually
  # provide a Cancel button (address book -> "Dial" -> Info-popup for
  # connecting; has Cancel button and should pop down when the connection is
  # established.

  # The info for the user is on a different axis than logging importance.
  # So it shouldn't be forced into it, but I'll do it anyway :)
  # difference is viewpoint: from User, not developer.

  # info for the developer, but with this mode engaged, you're probably also
  # the user. Can you spell tee?
  def debug(objects)
    objects.each { |obj|
      @@Logger.puts(" DEBUG #{obj.inspect}")
      @@Informer.status(obj.inspect)
    }
  end
  alias d debug

  # no info for the user
  def log(string)
    @@Logger.puts("  INFO #{string}")
  end

  # info for the user, status "bar"
  def report(string)
    @@Logger.puts("REPORT #{string}")
    @@Informer.status(string)
  end
  alias r report

  # info, but big step. Popup with Cancel button, log as INFO
  # *curses*
  # in order to be able to remove the popup, en event must be passed to it, or
  # it must be destroyed explicitely. That means taking it out of the log4r
  # context, which is a pity but not too bad. Hey, log4r is packaged for iPAQ
  # now :)
  def action(string, action_thread)
    @@Logger.puts(string)
    @@Informer.status(string)
    @@Informer.dialog(string, action_thread)
  end

  # popup, only OK; log as WARN
  def warn(string)
    @@Logger.puts("  WARN #{string}")
    #@@app.status(string)
    @@Informer.dialog(string)
  end

  # popup, Retry + Quit; log as ERROR
  def error(string, &retry_proc)
    @@Logger.error(string)
    @@app.status(string)
    knobs = {:Quit=>proc {
      @@logger.warn("User termination on error.")
      exit()
    }}
    if block_given?
      knobs += {:Retry=>proc {
        @@logger.warn("Retrying after error (on request)...")
        @@app.status("Retrying after error (on request)...")
        yield
      }}
    end
    @@app.dialog(parent, knobs + {:modal=>parent})
  end

  def crash()  # log as FATAL, popup (if managable), only OK; terminate
    @@logger.fatal(string)
    # don't bother with the status bar
    @@app.dialog(parent, :modal=>parent, :ok=>proc { exit() })
  end

  # Wise whould limit the amount of popups (modal windows only or .. ones too?)
  # as DoS preventor.

  # On an iPAQ, debug can hardly be expected to dump content on a file, but
  # Log4R provides network Outputters, too.

end
