# Author: Kero van Gelder
# Copyright: Kero van Gelder (the GNU Lesser General Public License applies)
# Since: 24 Sep 2001

require 'thread'

module Serialize
  @inputs = []
  @outputs = []
  @errors = []
  @bindings = {}

=begin
== Serialize::(un)set_input/output/error
(De)Associated a block with an input/output/error file descriptor, to be
serialized with other input/output/error fd's, i.e. the blocks associated with
those fd's are never executed simultaneously.

The block must be given and you must ensure the block will let the associated
stream be saturated, that is, input must be consumed, output buffers must be
filled. (Otherwise the loop in Serialize::serialize will never block anymore,
though it will not prevent other blocks from being called.)

When Serialize::serialize starts, it calls all bindings, to make sure any
pending I/O is dealt with; But there need not be any I/O, so your binding must
check first whether there is any I/O. Sorry about that :(
=end
  def Serialize::set_input(input, &block)
    raise "Must provide a block"  unless block
    @inputs << input  unless @bindings.fetch(input, nil)
    @bindings.store(input, block)
  end

  def Serialize::set_output(output, &block)
    raise "Must provide a block"  unless block
    @outputs << output  unless @bindings.fetch(output, nil)
    @bindings.store(output, block)
  end

  def Serialize::set_error(error, &block)
    raise "Must provide a block"  unless block
    @errors << error  unless @bindings.fetch(error, nil)
    @bindings.store(error, block)
  end

  def Serialize::unset_input(input)
    @inputs.delete(input)
    @bindings.delete(input)
  end

  def Serialize::unset_output(output)
    @outputs.delete(output)
    @bindings.delete(output)
  end

  def Serialize::unset_error(error)
    @errors.delete(error)
    @bindings.delete(error)
  end

  Queues = Mutex.new()

=begin
== Serialize::serialize
Start serializing events. Reentrant.

It is strongly suggested that once this loop is started, you only set and unset
file descriptors in the blocks invoked by this thread, not other threads.
=end
  def Serialize::serialize()
    return  if Queues.locked?
    Queues.synchronize {
      available = @inputs + @outputs + @errors
      loop {
	available.each { |fd|
	  @bindings.fetch(fd, nil).call
	}
	break  if @inputs.empty? and @outputs.empty? and @errors.empty?
	begin
	  available = select(@inputs, @outputs, @errors)
	  available.flatten!  if available
	rescue IOError
	  puts "Serialize::serialize caught: #{$!}"
	end
      }
    }
    puts "Serialize::serialize() ended (no fd's left)"  if $DEBUG
  end
end
